/*
    Copyright 2009 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <openssl/evp.h>    // libcrypto.a -lgdi32

typedef unsigned char   u8;
typedef unsigned int    u32;



#define VER         "0.1.1"



void partygaming_enum(u8 *keypath);
int partygaming_decrypt(u8 *mykeystr, u8 *buff, int buffsz);
u8 *base64_decode(u8 *data, int *size);
void std_err(void);
void wait_exit(int ret);



int main(int argc, char *argv[]) {
    int     len;
    u8      *pwd;

    fputs("\n"
        "PartyGaming password decrypter "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        partygaming_enum("Software\\PartyGaming");
        partygaming_enum("Software\\PartyGamingNet");
    } else {
        pwd = strdup(argv[1]);
        len = strlen(pwd);
        pwd = base64_decode(pwd, &len);
        len = partygaming_decrypt(NULL, pwd, len);
        if(len > 0) printf("  PASSWORD: %.*s\n", len, pwd);
    }

    wait_exit(0);
    return(0);
}



int partygaming_decrypt_verify(u8 *buff, int buffsz) {  // blah, unused
    EVP_ENCODE_CTX  ctx;
    int     len;

    EVP_DecodeInit(&ctx);
    EVP_DecodeUpdate(
        &ctx,
        buff, &buffsz,
        buff, buffsz);
    EVP_DecodeFinal(
        &ctx,
        buff + buffsz, &len);
    buffsz += len;
    return(buffsz);
}



int partygaming_decrypt(u8 *mykeystr, u8 *buff, int buffsz) {
    EVP_CIPHER_CTX  ctx;
    int     len;
    u8      key[64],
            iv[64],
            *keystr = "PartyPoker";

    if(mykeystr) keystr = mykeystr;

    if(EVP_BytesToKey(
        EVP_des_ede3_cbc(),
        EVP_md5(),
        "salu\0\0\0\0",
        keystr,
        strlen(keystr),
        1,
        key,
        iv) <= 0) return(-1);

    EVP_CIPHER_CTX_init(&ctx);

    if(EVP_CipherInit(
        &ctx,
        EVP_des_ede3_cbc(),
        key,
        iv,
        0) <= 0) return(-1);

    if(EVP_CipherUpdate(
        &ctx,
        buff, &buffsz,
        buff, buffsz) <= 0) return(-1);

    if(EVP_CipherFinal_ex(
        &ctx,
        buff + buffsz, &len) <= 0) return(-1);
    buffsz += len;

    EVP_CIPHER_CTX_cleanup(&ctx);
    return(buffsz);
}



void partygaming_enum(u8 *keypath) {
    HKEY    key,
            keyv;
    DWORD   parlen,
            vallen,
            type;
    int     len,
            k,
            x;
    u8      regbuff[512],
            par[128],
            val[512],
            *p,
            *s;

    if(RegOpenKeyEx(HKEY_CURRENT_USER, keypath, 0, KEY_READ, &key) != ERROR_SUCCESS) return;
    p = regbuff + sprintf(regbuff, "%s\\", keypath);

    for(k = 0;; k++) {
        len = sizeof(regbuff);
        if(RegEnumKeyEx(key, k, p, (void *)&len, NULL, NULL, NULL, NULL) != ERROR_SUCCESS) break;

        partygaming_enum(regbuff);

        if(RegOpenKeyEx(HKEY_CURRENT_USER, regbuff, 0, KEY_READ, &keyv) != ERROR_SUCCESS) continue;
        printf("\n- %s:\n", p);

        for(x = 0;; x++) {
            parlen = sizeof(par) - 1;
            vallen = sizeof(val) - 1;
            if(RegEnumValue(
                keyv,
                x,
                par,
                &parlen,
                NULL,
                &type,
                val,
                &vallen) != ERROR_SUCCESS) break;

            par[parlen] = 0;
            val[vallen] = 0;

            if(type != REG_BINARY) continue;    // because only binary are encrypted

            // for various reasons is necessary to try to decrypt first the string in base64 mode and then as is, so do NOT touch!
            len = vallen;
            s = base64_decode(val, &len);
            len = partygaming_decrypt(NULL, s, len);    // s is allocated so we can perform any operation on it
            if(len > 0) printf("  %-17s %.*s\n", par, len, s);
            free(s);
            if(len < 0) {
                len = partygaming_decrypt(NULL, val, vallen);
                if(len > 0) printf("  %-17s %.*s\n", par, len, val);
            }
        }
        RegCloseKey(keyv);
    }

    RegCloseKey(key);
}



u8 *base64_decode(u8 *data, int *size) {
    int     len,
            xlen,
            a   = 0,
            b   = 0,
            c,
            step;
    u8      *buff,
            *limit,
            *p;
    static const u8 base[128] = {   // supports also the Gamespy base64
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x3e,0x00,0x00,0x00,0x3f,
        0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,
        0x0f,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x3e,0x00,0x3f,0x00,0x00,
        0x00,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,
        0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,0x30,0x31,0x32,0x33,0x00,0x00,0x00,0x00,0x00
    };

    if(!size || (*size < 0)) {      // use size -1 for auto size!
        len = strlen(data);
    } else {
        len = *size;
    }
    xlen = ((len >> 2) * 3) + 1;    // NULL included in output for text
    buff = malloc(xlen);
    if(!buff) return(NULL);

    p = buff;
    limit = data + len;

    for(step = 0; /* data < limit */; step++) {
        do {
            if(data >= limit) {
                c = 0;
                break;
            }
            c = *data;
            data++;
            if((c == '=') || (c == '_')) {  // supports also the Gamespy base64
                c = 0;
                break;
            }
        } while(c && ((c <= ' ') || (c > 0x7f)));
        if(!c) break;

        switch(step & 3) {
            case 0: {
                a = base[c];
                break;
            }
            case 1: {
                b = base[c];
                *p++ = (a << 2)        | (b >> 4);
                break;
            }
            case 2: {
                a = base[c];
                *p++ = ((b & 15) << 4) | (a >> 2);
                break;
            }
            case 3: {
                *p++ = ((a & 3) << 6)  | base[c];
                break;
            }
        }
    }
    *p = 0;

    len = p - buff;
    if(size) *size = len;
    if((len + 1) != xlen) {             // save memory
        buff = realloc(buff, len + 1);  // NULL included!
        if(!buff) return(NULL);
    }
    return(buff);
}



void std_err(void) {
    perror("\nError");
    wait_exit(1);
}



void wait_exit(int ret) {
    fprintf(stderr, "\n  press RETURN to exit\n");
    fgetc(stdin);
    exit(ret);
}


